from dateutil.parser import parse
from enum import Enum
import pandas as pd


class ColType(Enum):
    string = 1
    date_time = 2
    integer = 3
    currency = 4
    decimal = 5
    boolean = 6


class IntChecker:
    my_type = ColType.integer

    def check(self, string):
        try:
            int(string)
            return True
        except Exception:
            return False


class DateChecker:
    my_type = ColType.date_time

    def check(self, string):
        try:
            parse(string)
            return True
        except Exception:
            return False


class FloatChecker:
    my_type = ColType.decimal

    def check(self, string):
        try:
            float(string)
            return True
        except Exception:
            return False


class BoolChecker:
    my_type = ColType.boolean

    def check(self, string):
        try:
            bool(string)
            return True
        except Exception:
            return False


class StringChecker:
    my_type = ColType.string

    def check(self, string):
        try:
            str(string)
            return True
        except Exception:
            return False


class ColProcessor:
    checkers = [BoolChecker(), IntChecker(), FloatChecker(),
                DateChecker(), StringChecker()]

    def __init__(self, uniques):
        self.uniques = uniques

    def get_type(self):
        outcomes = {}
        uniques = self.uniques
        for checker in ColProcessor.checkers:
            positives = 0.0
            for unique in uniques:
                if checker.check(unique):
                    positives += 1
            outcomes[checker.my_type] = positives / len(uniques)

        return outcomes


class MatrixProcessor:
    def __init__(self, matrix, headers):
        self.matrix = matrix
        self.headers = headers

    def process(self):
        pd_matrix = pd.DataFrame(self.matrix, columns=self.headers[:, 0])

        unique_value_dict = {}
        col_num = 0
        for column_name in pd_matrix.columns:
            column_mapping = {"actual": self.headers[col_num, 1]}

            uniques = pd_matrix[column_name].unique()
            col_processor = ColProcessor(uniques)
            col_type = col_processor.get_type()
            column_mapping["projection"] = col_type

            unique_value_dict[column_name] = column_mapping
            col_num += 1
        return unique_value_dict,  pd_matrix
        # print data_matrix[column_name].unique

        # data_matrix[column_name].map(unique_value_dict[column_name])



#from Simon.DataGenerator import *
#from Simon.Encoder import *
import pandas as pd
import numpy as np

import tensorflow as tf
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Dense, Activation, Flatten, Input, Dropout, MaxPooling1D, Convolution1D
from tensorflow.keras.layers import LSTM, Lambda, Masking, Embedding, TimeDistributed, BatchNormalization, concatenate
from tensorflow.keras.optimizers import SGD
from tensorflow.keras import backend as K
import tensorflow.keras.callbacks as callbacks

import re
import os
import time
import pickle
import logging

logger = logging.getLogger(__name__)

# record history of training
class LossHistory(callbacks.Callback):
    def on_train_begin(self, logs={}):
        self.losses = []
        self.accuracies = []

    def on_batch_end(self, batch, logs={}):
        self.losses.append(logs.get('loss'))
        self.accuracies.append(logs.get('binary_accuracy'))

class Simon:
    def __init__(self,encoder):
        self.encoder = encoder

    def binarize(self,x, sz=71):
        return tf.cast(tf.one_hot(x, sz, on_value=1, off_value=0, axis=-1), dtype=tf.float32)

    def clear_session(self):
        K.clear_session()

        return
    
    def custom_multi_label_accuracy(y_true, y_pred):
        # need some threshold-specific rounding code here, presently only for 0.5 thresh.
        return K.mean(K.round(np.multiply(y_true,y_pred)),axis=0)
    
    def eval_binary_accuracy(self,y_test, y_pred):
        correct_indices = y_test==y_pred
        all_correct_predictions = np.zeros(y_test.shape)
        all_correct_predictions[correct_indices] = 1
        return np.mean(all_correct_predictions),np.mean(all_correct_predictions, axis=0),all_correct_predictions
    
    def eval_confusion(self,y_test, y_pred):
        wrong_indices = y_test!=y_pred
        all_wrong_predictions = np.zeros(y_test.shape)
        all_wrong_predictions[wrong_indices] = 1
        return np.mean(all_wrong_predictions),np.mean(all_wrong_predictions, axis=0),all_wrong_predictions
    
    def eval_false_positives(self,y_test, y_pred):
        false_positive_matrix = np.zeros((y_test.shape[1],y_test.shape[1]))
        false_positives = np.multiply(y_pred,1-y_test)
        for i in np.arange(y_test.shape[0]):
            for j in np.arange(y_test.shape[1]) :
                if(false_positives[i,j]==1): #positive label for ith sample and jth predicted category
                    for k in np.arange(y_test.shape[1]):
                        if(y_test[i,k]==1): #positive label for ith sample and kth true category
                            false_positive_matrix[j,k] +=1
        return np.sum(false_positive_matrix),np.sum(false_positive_matrix, axis=0),false_positive_matrix
    
    def eval_ROC_metrics(self,y_test, y_pred):
        true_positive = y_pred*y_test
        true_negative = (1-y_pred)*(1-y_test)
        false_positive = y_pred*(1-y_test)
        false_negative = (1-y_pred)*y_test

        return np.sum(true_positive,axis=0),np.sum(true_negative,axis=0),np.sum(false_positive,axis=0),np.sum(false_negative,axis=0)
    
    
    def binarize_outshape(self,in_shape):
        return in_shape[0], in_shape[1], 71

    def max_1d(x):
        return K.max(x, axis=1)

    def striphtml(html):
        p = re.compile(r'<.*?>')
        return p.sub('', html)

    def clean(s):
        return re.sub(r'[^\x00-\x7f]', r'', s)

    def setup_test_sets(self,X, y):
        ids = np.arange(len(X))
        np.random.shuffle(ids)

        # shuffle
        X = X[ids]
        y = y[ids]

        train_end = int(X.shape[0] * .6)
        cross_validation_end = int(X.shape[0] * .3 + train_end)
        test_end = int(X.shape[0] * .1 + cross_validation_end)
    
        X_train = X[:train_end]
        X_cv_test = X[train_end:cross_validation_end]
        X_test = X[cross_validation_end:test_end]

        y_train = y[:train_end]
        y_cv_test = y[train_end:cross_validation_end]
        y_test = y[cross_validation_end:test_end]

        data = type('data_type', (object,), {'X_train' : X_train, 'X_cv_test': X_cv_test, 'X_test': X_test, 'y_train': y_train, 'y_cv_test': y_cv_test, 'y_test':y_test})
        return data

    def generate_model(self, max_len, max_cells, category_count, activation='sigmoid', training = False):
        filter_length = [1, 3, 3]
        nb_filter = [40, 200, 1000]
        pool_size = 2
        # document input
        document = Input(shape=(max_cells, max_len), dtype='int64')
        # sentence input
        in_sentence = Input(shape=(max_len,), dtype='int64')
        # char indices to one hot matrix, 1D sequence to 2D
        embedded = Lambda(self.binarize, output_shape=self.binarize_outshape)(in_sentence)
        # embedded: encodes sentence
        for i in range(len(nb_filter)):
            embedded = Convolution1D(nb_filter[i],
                                     filter_length[i],
                                     activation='relu',
                                     kernel_initializer='glorot_normal')(embedded)

            embedded = Dropout(0.1)(embedded)
            embedded = MaxPooling1D(pool_size=pool_size)(embedded)
        
        forward_sent = LSTM(256, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2)(embedded, training = training)
        backward_sent = LSTM(256, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2, go_backwards=True)(embedded, training = training)

        sent_encode = concatenate([forward_sent, backward_sent], axis=-1)
        sent_encode = Dropout(0.3)(sent_encode)
        # sentence encoder

        encoder = Model(inputs=in_sentence, outputs=sent_encode)

        #logger.debug(encoder.summary())
        encoded = TimeDistributed(encoder)(document)

        # encoded: sentences to bi-lstm for document encoding
        forwards = LSTM(128, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2)(encoded, training = training)
        backwards = LSTM(128, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2, go_backwards=True)(encoded, training = training)

        merged = concatenate([forwards, backwards], axis=-1)
        output = Dropout(0.3)(merged)
        output = Dense(128, activation='relu')(output)
        output = Dropout(0.3)(output)
        output = Dense(category_count, activation=activation)(output)
        # output = Activation('softmax')(output)
        model = Model(inputs=document, outputs=output)

        return model
        
    def generate_transfer_model(self,max_len, max_cells, category_count_prior, category_count_post, checkpoint, checkpoint_dir,activation='sigmoid'):
        filter_length = [1, 3, 3]
        nb_filter = [40, 200, 1000]
        pool_size = 2
        # document input
        document = Input(shape=(max_cells, max_len), dtype='int64')
        # sentence input
        in_sentence = Input(shape=(max_len,), dtype='int64')
        # char indices to one hot matrix, 1D sequence to 2D
        embedded = Lambda(self.binarize, output_shape=self.binarize_outshape)(in_sentence)
        # embedded: encodes sentence
        for i in range(len(nb_filter)):
            embedded = Convolution1D(nb_filter[i],
                                     filter_length[i],
                                     activation='relu',
                                     kernel_initializer='glorot_normal')(embedded)

            embedded = Dropout(0.1)(embedded)
            embedded = MaxPooling1D(pool_size=pool_size)(embedded)

        forward_sent = LSTM(256, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2)(embedded)
        backward_sent = LSTM(256, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2, go_backwards=True)(embedded)

        sent_encode = concatenate([forward_sent, backward_sent], axis=-1)
        sent_encode = Dropout(0.3)(sent_encode)
        # sentence encoder

        encoder = Model(inputs=in_sentence, outputs=sent_encode)

        #logger.debug(encoder.summary())
        encoded = TimeDistributed(encoder)(document)

        # encoded: sentences to bi-lstm for document encoding
        forwards = LSTM(128, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2)(encoded)
        backwards = LSTM(128, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2, go_backwards=True)(encoded)

        merged = concatenate([forwards, backwards], axis=-1)
        output_pre = Dropout(0.3)(merged)
        output_pre = Dense(128, activation='relu')(output_pre)
        output_pre = Dropout(0.3)(output_pre)
        output = Dense(category_count_prior, activation=activation)(output_pre)
        model = Model(inputs=document, outputs=output)
    
        # having built model of prior size, load weights
        self.load_weights(checkpoint, None, model, checkpoint_dir)
        # having loaded weights,rebuild model using new category_count in last layer
        output = Dense(category_count_post, activation='sigmoid')(output_pre)
        model = Model(inputs=document, outputs=output)
        # retrain the last layer
        for layer in model.layers[:8]:
            layer.trainable = False
        model.layers[8].trainable = True

        return model

    def generate_feature_model(self, max_len, max_cells, category_count, checkpoint_dir, config, DEBUG = True):
        '''
            Generate SIMON feature model which produces 128-d SIMON features as last layer
        '''
        filter_length = [1, 3, 3]
        nb_filter = [40, 200, 1000]
        pool_size = 2
        # document input
        document = Input(shape=(max_cells, max_len), dtype='int64')
        # sentence input
        in_sentence = Input(shape=(max_len,), dtype='int64')
        # char indices to one hot matrix, 1D sequence to 2D
        embedded = Lambda(self.binarize, output_shape=self.binarize_outshape)(in_sentence)
        # embedded: encodes sentence
        for i in range(len(nb_filter)):
            embedded = Convolution1D(nb_filter[i],
                                     filter_length[i],
                                     activation='relu',
                                     kernel_initializer='glorot_normal')(embedded)

            embedded = Dropout(0.1)(embedded)
            embedded = MaxPooling1D(pool_size=pool_size)(embedded)

        forward_sent = LSTM(256, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2)(embedded)
        backward_sent = LSTM(256, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2, go_backwards=True)(embedded)

        sent_encode = concatenate([forward_sent, backward_sent], axis=-1)
        sent_encode = Dropout(0.3)(sent_encode)
        # sentence encoder

        encoder = Model(inputs=in_sentence, outputs=sent_encode)

        #logger.debug(encoder.summary())
        encoded = TimeDistributed(encoder)(document)

        # encoded: sentences to bi-lstm for document encoding
        forwards = LSTM(128, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2)(encoded)
        backwards = LSTM(128, return_sequences=False, dropout=0.2,
                        recurrent_dropout=0.2, go_backwards=True)(encoded)

        merged = concatenate([forwards, backwards], axis=-1)
        output_pre = Dropout(0.3)(merged)
        output_pre = Dense(128, activation='relu', name='features')(output_pre)
        output = Dropout(0.3)(output_pre)
        output = Dense(category_count, activation='softmax')(output)
        model = Model(inputs=document, outputs=output)
        self.load_weights(config['checkpoint'],config,model,checkpoint_dir)

        '''GENERATE MODEL FOR INTERMEDIATE LAYER'''
        intermediate_model = Model(inputs=model.input, outputs=model.get_layer('features').output)
        return intermediate_model

    def plot_loss(self,history):
        import matplotlib.pyplot as plt
        # summarize history for accuracy
        plt.subplot('121')
        plt.plot(history.history['binary_accuracy'])
        plt.plot(history.history['val_binary_accuracy'])
        plt.title('model accuracy')
        plt.ylabel('accuracy')
        plt.xlabel('epoch')
        plt.legend(['train', 'test'], loc='upper left')
        # summarize history for loss
        plt.subplot('122')
        plt.plot(history.history['loss'])
        plt.plot(history.history['val_loss'])
        plt.title('model loss')
        plt.ylabel('loss')
        plt.xlabel('epoch')
        plt.legend(['train', 'test'], loc='upper left')
        plt.show()

    def train_model(self,batch_size, checkpoint_dir, model, nb_epoch, data):
        logger.debug("starting learning")
    
        check_cb = callbacks.ModelCheckpoint(checkpoint_dir + "text-class" + '.{epoch:02d}-{val_loss:.2f}.hdf5',
                                                    monitor='val_loss', verbose=0, save_best_only=True, mode='min')
        earlystop_cb = callbacks.EarlyStopping(monitor='val_loss', patience=7, verbose=1, mode='auto')

        tbCallBack = callbacks.TensorBoard(log_dir='./logs', histogram_freq=0, write_graph=True, write_images=False, embeddings_freq=0,
                                    embeddings_layer_names=None, embeddings_metadata=None)
        loss_history = LossHistory()
        history = model.fit(data.X_train, data.y_train, validation_data=(data.X_cv_test, data.y_cv_test), batch_size=batch_size,
                    nb_epoch=nb_epoch, shuffle=True, callbacks=[earlystop_cb, check_cb, loss_history, tbCallBack])
    
        logger.debug('losses: ')
        logger.debug(history.history['loss'])
        logger.debug('accuracies: ')
        logger.debug(history.history['val_binary_accuracy'])
        return history

    def evaluate_model(self,max_cells, model, data, encoder, p_threshold):
        logger.debug("Starting predictions:")
    
        start = time.time()
        scores = model.evaluate(data.X_test, data.y_test, verbose=0)
        end = time.time()
        logger.debug("Accuracy: %.2f%% \n Time: {0}s \n Time/example : {1}s/ex".format(
            end - start, (end - start) / data.X_test.shape[0]) % (scores[1] * 100))
        
        probabilities = model.predict(data.X_test, verbose=1)
        y_pred = np.where(y > p_threshold, 1, 0).astype(int)

        logger.debug("'Binary' accuracy ((TP+TN)/total) sample number is:")
        logger.debug(self.eval_binary_accuracy(data.y_test,y_pred)[0])
        logger.debug("'Binary' confusion ((FP+FN)/total) sample number is:")
        logger.debug(self.eval_confusion(data.y_test,y_pred)[0])
        logger.debug("False Positive sample number is:")
        logger.debug(self.eval_false_positives(data.y_test,y_pred)[0])
        TP,TN,FP,FN = self.eval_ROC_metrics(data.y_test, y_pred)
        logger.debug("Precision is:")
        logger.debug(np.sum(TP)/(np.sum(TP)+np.sum(FP)))
        logger.debug("Recall is:")
        logger.debug(np.sum(TP)/(np.sum(TP)+np.sum(FN)))
        logger.debug("F1 score is:")
        logger.debug(2*np.sum(TP)/(2*np.sum(TP)+np.sum(FP)+np.sum(FN)))


        return encoder.reverse_label_encode(probabilities,p_threshold)

    def resolve_file_path(self,filename, dir):
        if os.path.isfile(str(filename)):
            return str(filename)
        elif os.path.isfile(str(dir + str(filename))):
            return dir + str(filename)

    def load_weights(self,checkpoint_name, config, model, checkpoint_dir):
        if config and not checkpoint_name:
            checkpoint_name = config['checkpoint']
        if checkpoint_name:
            checkpoint_path = self.resolve_file_path(checkpoint_name, checkpoint_dir)
            #logger.debug("Checkpoint : %s" % str(checkpoint_path))
            model.load_weights(checkpoint_path)

    def save_config(self,execution_config, checkpoint_dir):
        filename = ""
        if not execution_config["checkpoint"] is None:
            filename = execution_config["checkpoint"].rsplit( ".", 1 )[ 0 ] + ".pkl"
        else:
            filename = time.strftime("%Y%m%d-%H%M%S") + ".pkl"
        with open(checkpoint_dir + filename, 'wb') as output:
            pickle.dump(execution_config, output, pickle.HIGHEST_PROTOCOL)

    def load_config(self,execution_config_path, dir):
        execution_config_path = self.resolve_file_path(execution_config_path, dir)
        config = pickle.load( open( execution_config_path, "rb" ) )
        self.encoder = config['encoder']
        return config

    def get_best_checkpoint(self,checkpoint_dir):
        max_mtime = 0
        max_file = ''
        for dirname,subdirs,files in os.walk(checkpoint_dir):
            for fname in files:
                full_path = os.path.join(dirname, fname)
                mtime = os.stat(full_path).st_mtime
                if mtime > max_mtime:
                    max_mtime = mtime
                    max_dir = dirname
                    max_file = fname
        return max_file

    ###### WHAT FOLLOWS IS FCHOLLET'S MULTI-GPU CODE #######
    def _get_available_devices(self):
        from tensorflow.python.client import device_lib
        local_device_protos = device_lib.list_local_devices()
        return [x.name for x in local_device_protos]


    def multi_gpu_model(self, model, gpus):
        """Replicates a model on different GPUs.
        Specifically, this function implements single-machine
        multi-GPU data parallelism. It works in the following way:
        - Divide the model's input(s) into multiple sub-batches.
        - Apply a model copy on each sub-batch. Every model copy
            is executed on a dedicated GPU.
        - Concatenate the results (on CPU) into one big batch.
        E.g. if your `batch_size` is 64 and you use `gpus=2`,
        then we will divide the input into 2 sub-batches of 32 samples,
        process each sub-batch on one GPU, then return the full
        batch of 64 processed samples.
        This induces quasi-linear speedup on up to 8 GPUs.
        This function is only available with the TensorFlow backend
        for the time being.
        # Arguments
            model: A Keras model instance. To avoid OOM errors,
                this model could have been built on CPU, for instance
                (see usage example below).
            gpus: Integer >= 2, number of on GPUs on which to create
                model replicas.
        # Returns
            A Keras `Model` instance which can be used just like the initial
            `model` argument, but which distributes its workload on multiple GPUs.
        # Example
        ```python
            import tensorflow as tf
            from keras.applications import Xception
            num_samples = 1000
            height = 224
            width = 224
            num_classes = 1000
            # Instantiate the base model
            # (here, we do it on CPU, which is optional).
            with tf.device('/cpu:0'):
                model = Xception(weights=None,
                                input_shape=(height, width, 3),
                                classes=num_classes)
            # Replicates the model on 8 GPUs.
            # This assumes that your machine has 8 available GPUs.
            parallel_model = multi_gpu_model(model, gpus=8)
            parallel_model.compile(loss='categorical_crossentropy',
                                optimizer='rmsprop')
            # Generate dummy data.
            x = np.random.random((num_samples, height, width, 3))
            y = np.random.random((num_samples, num_classes))
            # This `fit` call will be distributed on 8 GPUs.
            # Since the batch size is 256, each GPU will process 32 samples.
            parallel_model.fit(x, y, epochs=20, batch_size=256)
        ```
        """
        if K.backend() != 'tensorflow':
            raise ValueError('`multi_gpu_model` is only available '
                            'with the TensorFlow backend.')
        if gpus <= 1:
            raise ValueError('For multi-gpu usage to be effective, '
                            'call `multi_gpu_model` with `gpus >= 2`. '
                            'Received: `gpus=%d`' % gpus)

        target_devices = ['/device:CPU:0'] + ['/device:GPU:%d' % i for i in range(gpus)]
        available_devices = self._get_available_devices()
        for device in target_devices:
            if device not in available_devices:
                raise ValueError(
                    'To call `multi_gpu_model` with `gpus=%d`, '
                    'we expect the following devices to be available: %s. '
                    'However this machine only has: %s. '
                    'Try reducing `gpus`.' % (gpus,
                                            target_devices,
                                            available_devices))

        def get_slice(data, i, parts):
            shape = tf.shape(input=data)
            batch_size = shape[:1]
            input_shape = shape[1:]
            step = batch_size // parts
            if i == gpus - 1:
                size = batch_size - step * i
            else:
                size = step
            size = tf.concat([size, input_shape], axis=0)
            stride = tf.concat([step, input_shape * 0], axis=0)
            start = stride * i
            return tf.slice(data, start, size)

        all_outputs = []
        for i in range(len(model.outputs)):
            all_outputs.append([])

        # Place a copy of the model on each GPU,
        # each getting a slice of the inputs.
        for i in range(gpus):
            with tf.device('/device:GPU:%d' % i):
                with tf.compat.v1.name_scope('replica_%d' % i):
                    inputs = []
                    # Retrieve a slice of the input.
                    for x in model.inputs:
                        input_shape = tuple(x.get_shape().as_list())[1:]
                        slice_i = Lambda(get_slice,
                                        output_shape=input_shape,
                                        arguments={'i': i,
                                                    'parts': gpus})(x)
                        inputs.append(slice_i)

                    # Apply model on slice
                    # (creating a model replica on the target device).
                    outputs = model(inputs)
                    if not isinstance(outputs, list):
                        outputs = [outputs]

                    # Save the outputs for merging back together later.
                    for o in range(len(outputs)):
                        all_outputs[o].append(outputs[o])

        # Merge outputs on CPU.
        with tf.device('/device:CPU:0'):
            merged = []
            for outputs in all_outputs:
                merged.append(concatenate(outputs,
                                        axis=0))
            return Model(model.inputs, merged)

    def tune_ROC_metrics(self,max_cells, model, data, encoder,p_thresholds):
        logger.debug("Starting to compute ROC metrics...")

        start = time.time()
        scores = model.evaluate(data.X_test, data.y_test, verbose=0)
        end = time.time()
        logger.debug("Initial Accuracy: %.2f%% \n Time: {0}s \n Time/example : {1}s/ex".format(
            end - start, (end - start) / data.X_test.shape[0]) % (scores[1] * 100))

        probabilities = model.predict(data.X_test, verbose=1)

        Categories = encoder.categories
        logger.debug("The fixed categories are:")
        logger.debug(Categories)

        # evaluate ROC metrics for a number of p_threholds
        TPR_arr = np.zeros((p_thresholds.shape[0],1))
        FPR_arr = np.zeros((p_thresholds.shape[0],1))
        i = 0
        for p_threshold in p_thresholds:
            y_pred = np.where(y > p_threshold, 1, 0).astype(int)
            TP,TN,FP,FN = self.eval_ROC_metrics(data.y_test, y_pred)
            TPR_arr[i,:]= np.sum(TP)/(np.sum(TP)+np.sum(FN))
            FPR_arr[i,:]= np.sum(FP)/(np.sum(FP)+np.sum(TN))
            i = i+1

        # save data for analysis later?
        # np.save("TP",TP)
        # np.save("TN",TN)
        # np.save("FP",FP)
        # np.save("FN",FN)
        # np.save("TPR_arr",TPR_arr)
        # np.save("FPR_arr",FPR_arr)

        return TPR_arr,FPR_arr



from flask import Flask, request
from json import JSONEncoder
import pandas
import pickle
import numpy as np
import configparser
import time

from Simon import *
from Simon.Encoder import *
from Simon.DataGenerator import *
from Simon.LengthStandardizer import *

from Simon.penny.guesser import guess

class SimonRestListener:
    """ SimonRestListener accepts a pandas dataframe (or a path to one),
    uses its predictive model to generate labels for the columns,
    and then encodes this in JSON to be returned to the caller
    """
    def __init__(self, modelName):
       self.model =  modelName
       self.encoder = JSONEncoder()

    def runModel(self, frame, p_threshold):

        # setup model as you typically would in a Simon main file
        maxlen = 20
        max_cells = 500
        p_threshold = 0.5

        DEBUG = True # boolean to specify whether or not print DEBUG information

        checkpoint_dir = "/clusterfiles/scripts/pretrained_models/"

        with open('/clusterfiles/scripts/Categories.txt','r') as f:
            Categories = f.read().splitlines()

        # orient the user a bit
        print("fixed categories are: ")
        Categories = sorted(Categories)
        print(Categories)
        category_count = len(Categories)

        execution_config=modelName

        # load specified execution configuration
        if execution_config is None:
            raise TypeError
        Classifier = Simon(encoder={}) # dummy text classifier

        config = Classifier.load_config(execution_config, checkpoint_dir)
        encoder = config['encoder']
        checkpoint = config['checkpoint']

        X = encoder.encodeDataFrame(frame)
            
        # build classifier model
        model = Classifier.generate_model(maxlen, max_cells, category_count)
        Classifier.load_weights(checkpoint, None, model, checkpoint_dir)
        model_compile = lambda m: m.compile(loss='binary_crossentropy', optimizer='adam', metrics=['binary_accuracy'])
        model_compile(model)
        y = model.predict(X)
        # discard empty column edge case
        y[np.all(frame.isnull(),axis=0)]=0

        result = encoder.reverse_label_encode(y,p_threshold)

        ## LABEL COMBINED DATA AS CATEGORICAL/ORDINAL
        print("Beginning Guessing categorical/ordinal classifications...")
        start_time_guess = time.time()
        category_count = 0
        ordinal_count = 0
        raw_data = frame.as_matrix()
        for i in np.arange(raw_data.shape[1]):
            tmp = guess(raw_data[:,i], for_types ='category')
            if tmp[0]=='category':
                category_count += 1
                tmp2 = list(result[0][i])
                tmp2.append('categorical')
                result[0][i] = tuple(tmp2)
                result[1][i].append(1)
                if ('int' in result[1][i]) or ('float' in result[1][i]) \
                    or ('datetime' in result[1][i]):
                        ordinal_count += 1
                        tmp2 = list(result[0][i])
                        tmp2.append('ordinal')
                        result[0][i] = tuple(tmp2)
                        result[1][i].append(1)
        elapsed_time = time.time()-start_time_guess
        print("Total statistical variable guessing time is : %.2f sec" % elapsed_time)
        ## FINISHED LABELING COMBINED DATA AS CATEGORICAL/ORDINAL

        Classifier.clear_session()
        
        return self.encoder.encode((result))

    def predict(self, request_data,p_threshold):
        frame = pickle.loads(request_data)
                
        return self.runModel(frame,p_threshold)

    def predictFile(self, fileName,p_threshold):
        frame = pandas.read_csv(str(fileName),dtype='str')
        return self.runModel(frame,p_threshold)
        
config = configparser.ConfigParser()
config.read('config.ini')
modelName = config['DEFAULT']['modelName']

# modelName = "Base.pkl" # consider specifying explicitly before building each docker image
print("using model " + modelName + " ...")
        
listener = SimonRestListener(modelName)

app = Flask(__name__)

@app.route("/", methods=['POST'])
def predict():
    """ Listen for data being POSTed on root. The data expected to 
    be a string representation of a pickled pandas frame
    """
    request.get_data()
    p_threshold=0.5 # this will become class-dependent and tunable shortly (backwards compatible)
    return listener.predict(request.data,p_threshold)


@app.route("/fileName", methods=['POST'])
def predictFile():
    """ Listen for data being POSTed on root. The data should 
    be a string representation of a file name accessible to flask server
    """
    request.get_data()
    p_threshold=0.5 # this will become class-dependent and tunable shortly (backwards compatible)
    return listener.predictFile(request.data.decode("utf-8"),p_threshold)
    
@app.route("/fileUpload", methods=['POST'])
def predictUploadedFile():
    """ Listen for data being POSTed on root. The data should 
    be a string representation of a pickled numpy array
    """
    request.get_data()
    file = request.files['file']
    fileName = '/clusterfiles/uploaded_file.csv'
    file.save(fileName)
    p_threshold=0.5 # this will become class-dependent and tunable shortly (backwards compatible)
    result = listener.predictFile(fileName,p_threshold)
    os.remove(fileName)
    
    return result

# $env:FLASK_APP="rest/SimonRestListener.py"
# flask run


from datetime import datetime

def to_unicode(w):
    if isinstance(w, str):
        return w
#    if isinstance(w, basestring):
#        return str(w, errors='ignore')
    else:
        return str(w)

def prep_value(w):
    return to_unicode(w).strip()


def strip_non_ascii(string):
    ''' Returns the string without non ASCII characters'''
    if type(string) == datetime:
        return str(string)
    stripped = (c for c in string if 0 < ord(c) < 127)
    return ''.join(stripped)



from faker import Factory
import numpy as np
import pandas as pd
import pprint
import json
from Simon.ColumnProcessor import ColType as ct
import Simon.ColumnProcessor as cp
import os.path
from random import randint
import random


class FakeDataCreator:
    methods = None
    with open(os.path.join(os.path.dirname(__file__),'types.json')) as data_file:
        methods = json.load(data_file)
    filtered_dict = {k: v for k, v in methods.items() if len(v)
                     > 0 and not v in ['']}

    def __init__(self):
        fake = Factory.create()
        fake.random.seed(1234)
        self.fake = fake

    def get_generator(self):
        fake_func = None
        column_name = ""
        valid = False
        fake = self.fake
        column_category = ""
        while not valid:
            try:
                #func_idx = randint(dir(fake).index('address'), len(dir(fake)))
                #column_name = dir(fake)[func_idx]
                column_name = random.choice(
                    list(FakeDataCreator.filtered_dict.keys()))
                fake_func = getattr(fake, column_name)
                valid = (not column_name in ['binary']) and callable(fake_func) and not isinstance(
                    fake_func(), (list, tuple, np.ndarray)) and str(fake_func())
                #print("{0} - {1}".format(column_name, str(fake_func())))
            except Exception as ex:
                print (ex)
                continue
        return fake_func, column_name

    def map_column_names_to_types(self, column_names):
        # print(column_names)
        filtered_dict = FakeDataCreator.filtered_dict
        return np.array([filtered_dict[column_name] if (column_name in filtered_dict) else None for column_name in column_names])


class DataGenerator:
    # this method isn't getting used currently..  it tries to create a fixed
    # dataset and describe it.
    # this may be beneficial to help us create predictable results and while
    # doing data encrichment
    @staticmethod
    def gen_data(count):
        path = "data/"
        matrix_name = path + "matrix_{0}.npy".format(count)
        header_name = path + "header_{0}.npy".format(count)
        if os.path.isfile(matrix_name) and os.path.isfile(header_name):
            print("loading data")
            return np.load(matrix_name), np.load(header_name)
        print("creating data")
        fake = Factory.create()
        fake.random.seed(1234)

        data = []
        for _ in range(0, count):
            first = fake.first_name()
            last = fake.last_name()
            address = fake.address()
            date = fake.date_time().strftime("%c")
            is_active = fake.boolean()
            cc_num = fake.credit_card_number()
            phone = fake.phone_number()
            ssn = fake.ssn()
            email = fake.safe_email()
            balance = fake.random_number()
            data.append([first, last, last + ", " + first, address,
                         date, is_active, cc_num, phone, ssn, email, balance])

        headers = np.array([["FIRST", ct.string], ["LAST_NAME", ct.string], ["NAME", ct.string], ["ADDRESS", ct.string], ["DATE", ct.date_time], [
                           "IsActive", ct.boolean], ["CC#", ct.string], ["Phone", ct.string], ["SS#", ct.string], ["Email", ct.string], ["Balance", ct.decimal]])
        #headers = headers[np.newaxis,:]
        matrix = np.matrix(data)
        with open(matrix_name, 'wb') as matrix_file:
            np.save(matrix_file, matrix)
        with open(header_name, 'wb') as header_file:
            np.save(header_file, headers)

        return matrix, headers

    @staticmethod
    def gen_col_data(data_creator, rows):
        fake_func, column_name = data_creator.get_generator()
        #print ("fake: {0}, col: {1}".format(fake_func, column_name))
        ret_val = np.empty((rows,), dtype="object")
        for i in range(rows):
            try:
                val = fake_func()
                # TODO: utf-8 is probably not acceptable for
                # internationalization
                ret_val[i] = str(val)  # repr.encode('utf-8')
            except Exception:
                print("error: {0}".format(fake_func()))
        return ret_val, column_name
    # generates a dataset given a certain size
    # will save it to data/ when complete

    @staticmethod
    def gen_test_data(shape, try_reuse_data):
        str_shape = 'x'.join([str(x) for x in shape])
        data_path = "data/"
        if not os.path.isdir(data_path):
            os.makedirs(data_path)
        matrix_name = data_path + "matrix_{0}.npy".format(str_shape)
        header_name = data_path + "header_{0}.npy".format(str_shape)
        data_creator = FakeDataCreator()
        if try_reuse_data and os.path.isfile(matrix_name) and os.path.isfile(header_name):
            print("loading data")
            ret_val = np.load(matrix_name)
            temp_header = data_creator.map_column_names_to_types(np.load(header_name))
            ret_header = np.asarray([temp_header[i] for i in np.arange(temp_header.shape[0]) if not temp_header[i]==None])
            # print("ret_header shape:")
            # print(ret_header.shape)
            del_idx = np.zeros(ret_val.shape[1]-ret_header.shape[0], dtype='object')
            k = 0
            for j in np.arange(temp_header.shape[0]):
                if(temp_header[j]==None):
                    # print("deleting column %d"%j)
                    del_idx[k] = j
                    k = k+1
            # print("del_idx:")
            # print(del_idx)
            ret_val = np.delete(ret_val,del_idx,axis=1) #delete redundant faker columns from data...
            # print("ret_header")
            # print(ret_header)
            # print(ret_header.shape)
            # print(ret_val.shape)
            return ret_val, ret_header 
        print("creating data")
        ret_val = np.empty(shape, dtype="object")
        headers = np.empty(shape[1], dtype="object")

        for j in range(shape[1]):
            col_data, column_name = DataGenerator.gen_col_data(data_creator, shape[0])
            # fake_func, column_name = data_creator.get_generator()
            headers[j] = column_name
            ret_val[:, j] = col_data
            # for i in range(shape[0]):
            #     try:
            #         val= fake_func()
            #         #TODO: utf-8 is probably not acceptable for
            #         internationalization
            #         ret_val[i,j] = str(val)#repr.encode('utf-8')
            #     except Exception:
            #         print "error: {0}".format(fake_func())

        with open(matrix_name, 'wb') as matrix_file:
            np.save(matrix_file, ret_val)
        with open(header_name, 'wb') as header_file:
            np.save(header_file, headers)
        
        temp_header = data_creator.map_column_names_to_types(headers)
        ret_header = np.asarray([temp_header[i] for i in np.arange(temp_header.shape[0]) if not temp_header[i]==None])
        del_idx = np.zeros(ret_val.shape[1]-ret_header.shape[0], dtype='object')
        k = 0
        for j in np.arange(temp_header.shape[0]):
            if(temp_header[j]==None):
                del_idx[k] = j #deleting column j
                k = k+1
        ret_val = np.delete(ret_val,del_idx,axis=1) #delete redundant faker columns from data...
        
        return ret_val, ret_header

    @staticmethod
    def add_col_nulls(data, col_idx, percent=.05):
        count = int(data.shape[0] * percent)
        rows = [randint(0, data.shape[0] - 1) for p in range(count)]
        nulls = ['NULL', 'N/A', 'n/a', 'na', 'None', 'Nothing', '', '-']
        data[rows, col_idx] = random.choice(nulls)

    @staticmethod
    def add_nulls_uniform(data, percent=.05):
        for idx in range(data.shape[1]):
            DataGenerator.add_col_nulls(data,idx, percent)


if __name__ == '__main__':
    from Encoder import *
    import ColumnProcessor as cp
    data, header = DataGenerator.gen_test_data((100, 50))
    encoded_data = Encoder.encode_matrix(data, StringToIntArrayEncoder())

    processor = cp.MatrixProcessor(data, header)
    col_types, matrix = processor.process()
    print(matrix)
    #pd_matrix = pd.DataFrame(data, columns=header[:,0])
    #type_provider = cp.ColProcessor(pd_matrix)



import json
import pandas
import os.path
import numpy as np
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import MultiLabelBinarizer
from Simon.LengthStandardizer import *
import logging

class StringToIntArrayEncoder:
    def encode(self, string, char_count):
        return [ord(c) for c in string.ljust(char_count)[:char_count]]


class Encoder:
    def __init__(self,categories):
        self._char_indicies = {}
        self._indices_char = {}
        self.cur_max_cells = 0
        self._encoder = LabelEncoder()
        self._multi_encoder = MultiLabelBinarizer()
        self.categories = categories


    def process(self, raw_data, max_cells):
        chars = {}
        for doc in raw_data:
            for cell in doc:
                for c in str(cell):
                    if not c in chars:
                        chars[c] = True
        #this line is sounding increasingly unnecesary, probably should just always leave it capped to 500
        self.cur_max_cells = min(raw_data.shape[1], max_cells)
        chars = sorted(list(chars))

        self.char_indices = dict((c, i) for i, c in enumerate(chars))
        self._indices_char = dict((i, c) for i, c in enumerate(chars))
  
    def encode_matrix(self, matrix):
        chars = 20
        new_matrix = np.zeros((matrix.shape[0], matrix.shape[1], chars))
        for row in range(0, matrix.shape[0]):
            new_array = []
            for col in range(0, matrix.shape[1]):
                new_matrix[row, col] = self._encoder.encode(matrix[row, col], chars)
                
        return new_matrix

    def encode_data(self, raw_data, header, max_len):
        
        X = np.ones((raw_data.shape[0], self.cur_max_cells,
                     max_len), dtype=np.int64) * -1
        y = self.label_encode(header)
        
        # track unencoded chars
        unencoded_chars = None
        if not os.path.isfile('unencoded_chars.json'):
            open('unencoded_chars.json','w').close()
            unencoded_dict = {}
        else:
            with open('unencoded_chars.json') as data_file:
                unencoded_chars = json.load(data_file)
            unencoded_dict = {k: v for k, v in unencoded_chars.items() if not v in ['']}

        for i, column in enumerate(raw_data):
            for j, cell in enumerate(column):
                if(j < self.cur_max_cells):
                    string = cell[-max_len:].lower()
                    for k, char in enumerate(string):
                        try:
                            X[i, j, (max_len - 1 - k)] = self.char_indices[char]
                        except Exception:
                            #X[i, j, (max_len - 1 - k)] = self.char_indices[' ']
                            if char in unencoded_dict.keys():
                                unencoded_dict[char] = unencoded_dict[char]+1
                            else:
                                unencoded_dict[char] = 1
        with open('unencoded_chars.json','w') as out_file:
            json.dump(unencoded_dict, out_file)
        print("X shape: {0}\ny shape: {1}".format(X.shape, y.shape))
        
        return X, y

    def decode_matrix(self, X):
        ret_val = np.empty((X.shape[0], X.shape[1]), dtype='object')
        for i in range(X.shape[0]):
            for j in range(X.shape[1]):
                string = ""
                # start at the end since we reversed it
                for k in range(X.shape[2] - 1, -1, -1):
                    if X[i, j, k] < 0:
                        continue
                    string = string + self._indices_char[X[i, j, k]]
                ret_val[i, j] = string
        return ret_val

    def label_encode(self, Y):
        # encode class values as integers
        categories = self.categories
        self._multi_encoder.fit([categories])

        # convert integers to dummy variables (i.e.  one hot encoded)
        multi_encoder = self._multi_encoder
                
        ret_val = multi_encoder.transform(Y)
        
        return ret_val

    def reverse_label_encode(self, y, p_threshold):
        
        multi_encoder = self._multi_encoder
        
        # convert predictions to binary integers
        y_pred = np.where(y > p_threshold, 1, 0).astype(int)
        
        # extract label prediction probabilities
        label_probs = [[prob for prob, pred in zip(prob_list, pred_list) if pred == 1] 
            for prob_list, pred_list in zip(np.array(y), y_pred)]

        # extract labels
        labels = multi_encoder.inverse_transform(y_pred)

        return labels, label_probs
        
    def encodeDataFrame(self, df: pandas.DataFrame):
        
        nx,ny = df.shape
        
        out = DataLengthStandardizerRaw(df,self.cur_max_cells)
        
        raw_data = np.char.lower(np.transpose(out).astype('U'))
        
        return self.x_encode(raw_data, 20)  
        
        
    def x_encode(self, raw_data, max_len):

        X = np.ones((raw_data.shape[0], self.cur_max_cells,
                     max_len), dtype=np.int64) * -1        
                
        # track unencoded chars
        unencoded_chars = None
        if not os.path.isfile('unencoded_chars.json'):
            unencoded_dict = {}
        else:
            with open('unencoded_chars.json') as data_file:
                unencoded_chars = json.load(data_file)
            unencoded_dict = {k: v for k, v in unencoded_chars.items() if not v in ['']}

        for i, column in enumerate(raw_data):
            for j, cell in enumerate(column):
                if(j < self.cur_max_cells):
                    string = cell[-max_len:].lower()
                    for k, char in enumerate(string):
                        try:
                            X[i, j, (max_len - 1 - k)] = self.char_indices[char]
                        except Exception:
                            X[i, j, (max_len - 1 - k)] = self.char_indices[' '];
                            if char in unencoded_dict.keys():
                                unencoded_dict[char] = unencoded_dict[char]+1
                            else:
                                unencoded_dict[char] = 1
        with open('unencoded_chars.json','w') as out_file:
            json.dump(unencoded_dict, out_file)
        
        return X


# -*- coding: utf-8 -*-

from dateutil.parser import parse
#from penny.geo_lookup import get_places_by_type
#from address import AddressParser
from email.utils import parseaddr
from .utils import prep_value, strip_non_ascii
# import phonenumbers
import re
import time
import datetime
import os

cur_dir = os.path.dirname(os.path.realpath(__file__))


DELIMITERS = [',','|','/']


def is_a_nothing(value, header=None):
    value = prep_value(value).lower()

    if not value:
        return True
    
    if value in ['none','nan','null','n/a']:
        return True
    
    return False


def is_a_time(value, header=None):
    value = prep_value(value).lower()

    try:
        time.strptime(value, '%H:%M')
        return True
    except ValueError:
        try:
            time.strptime(value, '%H:%M %p')
            return True
        except ValueError:
            try:
                time.strptime(value, '%H:%M:%S')
                return True
            except ValueError:
                return False


def is_a_date(value, header=None):
    """
    Dateutil recognizes some letter combinations as dates, which is almost 
    always something that isn't really a date (like a US state abbreviation)
    """
    value = prep_value(value)

    if len(value) < 4:
        return False

    if value.startswith('-'):
        return False

    if not any(char.isdigit() for char in value):
        return False

    if is_a_time(value, header=header):
        return False

    """
    This is iffy. Obviously it's totally possible to have infinitely 
    precise measurements, but we're going to guess that if there are 
    more numbers to the right of the decimal point than the left, we 
    are probably dealing with a coordinate (or something)
    """
    if '.' in value:
        pieces = value.split('.')
        if len(pieces[1]) > len(pieces[0]):
            return False
        elif len(pieces) == 2 and len(pieces[0]) < 9:
            return False

    now = datetime.datetime.now()
    
    try:
        d = parse(value)
        # Hard to imagine a dataset only in the future
        if d.year < now.year + 5 and d.year > now.year - 300:
            return True
        return False
    except Exception as e:
        try:
            d = datetime.datetime.fromtimestamp(int(value))
            if d.year < now.year + 5 and d.year > now.year - 5:
                return True
        except:
            return False

        return False



def is_a_currency(value, header=None):
    value = prep_value(value)

    if value.startswith((u'$',u'£',u'€')):
        return True

    return False


def is_a_percent(value, header=None):
    value = prep_value(value)
    
    if value.endswith('%'):
        return True

    return False


def is_a_number(value, header=None):
    value = prep_value(value)
    pieces = value.split(',')
    if len(pieces) > 1 and not any([len(piece) == 3 for piece in pieces]):
        return False

    value = value.replace(',','')

    if header:
        header = prep_value(header).lower()

    if header and header in ['zip', 'zipcode', 'postal code']:
        return False

    if is_a_percent(value) or is_a_currency(value):
        return True

    if "." in value:
        try:
            v = float(value)
            return True
        except:
            return False
    else:
        try:
            int(value)
            return True
        except:
            return False


def is_a_text(value, header=None):
    value = prep_value(value).strip()

    return len(value) > 90


def is_a_label(value, header=None):
    value = prep_value(value)

    if not value or value == "":
        return False

    if is_a_number(value):
        return False

    if len(value) > 40:
        return False

    if '.' in value:
        return False

    if len(value.split(' ')) > 5:
        return False

    possible_matches = [d for d in DELIMITERS if d in value]
    
    # more than one delimeter is probably a sentence
    if len(possible_matches) > 1:
        return False

    return True


def is_a_latitude(value, header=None):
    value = prep_value(value)

    if header:
        header = unicode(header).lower().strip()

    if not is_a_number(value):
        return False

    if not abs(float(value)) <= 90:
        return False

    # so we know we have a value that is between -90 and 90
    header_names = ['lat', 'lt']
    if header and any([h in header for h in header_names]):
        return True

    return False


def is_a_longitude(value, header=None):
    value = prep_value(value)

    if header:
        header = unicode(header).lower().strip()

    if not is_a_number(value):
        return False

    if not abs(float(value)) <= 180:
        return False

    # so we know we have a value that is between -180 and 180
    header_names = ['lon', 'lng']
    if header and any([h in header for h in header_names]):
        return True

    return False


def is_a_coord_pair(value, header=None, pos=None):
    value = prep_value(value)

    disallowed = ['(',')','{','}','[',']']

    possible_matches = [d for d in DELIMITERS if d in value]
    
    # if more than one of these is present or none of them, than this isn't 
    # a pair
    if len(possible_matches) != 1:
        return False

    # Get rid of weird shit people put in coord pair columns
    for d in disallowed:
        value = value.replace(d,'')

    delimiter = possible_matches[0]
    possible_cords = value.split(delimiter)
    
    if len(possible_cords) != 2:
        return False

    # All parts have to be floats or ints
    if not all([is_a_number(x) for x in possible_cords]):
        return False

    # max abs lat is 90, max abs lng is 180
    if any([abs(float(x)) > 180 for x in possible_cords]):
        return False

    if all([abs(float(x)) > 90 for x in possible_cords]):
        return False

    # We have a pair of numbers, each of which is less than 180 and one of 
    # which is less than 90
    return True
    

def is_a_place(value, place_type, header=None):
    value = prep_value(value)

    if not is_a_label(value):
        return False

    # If your country's name is longer than 40 characters, you're doing 
    # something wrong.
    if len(value) > 40:
        return False

    non_addrs = ['|','/','?','!','@','$','%']
    if len([na for na in non_addrs if na in value]) > 0:
        return False

    if header:
        header = header.lower().strip()

    if header and header in place_type:
        return True

    if len(get_places_by_type(value, place_type)) > 0:
        return True

    if place_type in ['region'] and len(value) < 4 and \
        len(get_places_by_type(value, place_type + '_iso_code')) > 0:
        return True

    return False


def is_a_city(value, header=None, pos=None):
    return is_a_place(value, 'city', header=header)


def is_a_region(value, header=None, pos=None):
    return is_a_place(value, 'region', header=header)


def is_a_country(value, header=None, pos=None):
    return is_a_place(value, 'country', header=header)


state_list = [l.strip().lower() for l in open(cur_dir + '/data/us_states.txt').readlines()]
def is_a_state(value, header=None):
    value = prep_value(value)
    if value.lower() in state_list:
        return True
    
    return False


def is_a_zip(value, header=None, pos=None):
    value = prep_value(value)

    if header:
        header = prep_value(header).lower()

    if header and header in ['zip', 'zipcode', 'postal code']:
        return True

    if value.count('-') == 1 and len(value) == 10:
        primary = value.split('-')[0]
    else:
        primary = value

    try:
        primary = int(primary)
    except:
        return False

    primary = prep_value(primary)
    if len(primary) == 5 and int(primary) > 499:
        return True

    return False


ap = AddressParser()
def address_pieces(value):
    value = strip_non_ascii(prep_value(value))
    
    if is_a_number(value):
        return [], None

    if len(value) > 80:
        return [], None

    address = ap.parse_address(value)

    keys = [
        'house_number', 
        'street', 
        'city',
        'zip',
        'state'
    ]

    return [key for key in keys if getattr(address, key, None)], address


"""Check if a string is a house number + street name. The street part of an 
address. Note that we return false if this is a more complete address. """
def is_a_street(value, header=None, pos=None):
    has,address = address_pieces(value)

    if len(has) == 2 and 'house_number' in has and 'street' in has:
        return not is_a_address(value, header=header)
    
    return False


"""Check to see if this is enough of an address that it could be geocoded. So 
has at least a city + state, or at least street + city"""
def is_a_address(value, header=None, pos=None):
    value = prep_value(value)
    
    has,address = address_pieces(value)
    
    if len(has) >= 2:
        if getattr(address, 'city', None) is not None:
            return True
        
        pieces = value.split(' ')
        if len(pieces) > 2:
            """Sometimes we get an address like 100 Congress Austin TX, so let's 
            take a stab at breaking up the city/state in a way AddressParser 
            might understand"""
            pieces[len(pieces) - 2] = pieces[len(pieces) - 2] + ','

            has,address = address_pieces(' '.join(pieces))
            if len(has) > 2 and getattr(address, 'city', None):
                return True

    return False


def is_a_phone(value, header=None, pos=None):
    value = prep_value(value)

    if len(value) > 20:
        return False

    """Check for international numbers"""
    if value.startswith('+'):
        try:
            phonenumbers.parse(value)
            return True
        except:
            return False

    """Otherwise let's hope it's a US number"""
    reg = re.compile(".*?(\(?\d{3}\D{0,3}\d{3}\D{0,3}\d{4}).*?", re.S)
    matches = reg.search(value)

    """We're not looking for text fields that contain phone numbers, only fields 
    that are dedicated to phone number"""
    if matches and len(matches.group(1)) == len(value):
        return True


    return False


def is_a_email(value, header=None, pos=None):
    value = prep_value(value)

    possible = parseaddr(value)
    if possible[1] == '':
        return False
    
    e = re.compile(r'[\w\-][\w\-\.]+@[\w\-][\w\-\.]+[a-zA-Z]{1,4}')
    m = e.search(possible[1])
    if not m:
        return False

    if len(m.group(0)) == len(possible[1]):
        return True

    return False


def is_a_url(value, header=None, pos=None):
    value = prep_value(value)

    # blatantly ripped from Django
    regex = re.compile(
        r'(^https?://)?'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+[A-Z]{2,6}\.?|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    m = regex.search(value)

    if not m:
        return False

    return len(m.group(0)) == len(value)



from collections import Counter
from .utils import to_unicode
''' from .value_checks import (is_a_date, is_a_number, is_a_nothing,
    is_a_latitude, is_a_longitude, is_a_coord_pair, is_a_country, is_a_city, 
    is_a_state, is_a_address, is_a_text, is_a_label, is_a_zip, is_a_street, 
    is_a_phone, is_a_url, is_a_email, is_a_time, is_a_currency, is_a_percent) '''

# currently understands
# category
# datetime
# time
# number
# label
# text
# id
# email
# url
# address
# street
# city
# state
# zipcode
# country
# phone
# latitude
# longitude
# coordinate_pair

# coming soon
# name

# ordinal??? -- can obtain from categorical/int info...

from .utils import prep_value

def is_a_nothing(value, header=None):
    value = prep_value(value).lower()

    if not value:
        return True
    
    if value in ['none','nan','null','n/a']:
        return True
    
    return False


def guess(column_values, header=None, for_types=None):
    types = []
    checked_types = []
    threshold = .8

    def should_check(column_type):
        if not for_types or column_type in for_types:
            return True

        return False

    # Check if all values are unique
    if should_check('rowlabel'):
        if len(list(set(column_values))) == len(column_values):
            types.append('rowlabel')

    # Get non-empty values
    values = [v for v in column_values if not is_a_nothing(v)]
    count_not_empty = len(values)
    
    # If the column is empty, we don't need to do anything else
    if count_not_empty == 0:
        types.append('unknown')
        return sorted(list(set(types)))
    
    
    def do_check(test_func):
        passes_test_count = sum([test_func(v, header=header) for v in values])
        return  float(passes_test_count) / count_not_empty > threshold
    
    
    # if the column is long text, we don't need to do anything else
    if should_check('text') and do_check(is_a_text):
        types.append('text')
        return sorted(list(set(types)))


    # If the column is a date, we don't need to do anything else
    if should_check('datetime') and do_check(is_a_date):
        types.append('datetime')
        return sorted(list(set(types)))

    word_counts = Counter(values).items()
    largest_category_ratio = float(max([w[1] for w in word_counts])) / len(values)
    unique_value_ratio = float(len(word_counts)) / len(values)
    dot_ratio = float(sum([u'.' in to_unicode(v) for v in values])) / len(values)
    
    

    # periods are important to check for determining if numeric is a category
    has_dots = dot_ratio > threshold
    is_boolean = len(word_counts) == 2
    
    # See if values repeat often enough that we should count groups
    if should_check('category'):
        if is_boolean or (largest_category_ratio >= .05 and \
            unique_value_ratio < .2 and not has_dots and len(word_counts)<50):
            
            #print(largest_category_ratio)
            #print(unique_value_ratio)
            #print(word_counts)
            
            types.append('category')


    # Check if this is some kind of ID
    id_labels = ['_id', '_ID', '-id', '-ID', 'ID', ' ID', ' id']
    is_id_header = header and any([s in header for s in id_labels])

    
    if ('rowlabel' in types or 'category' in types) and is_id_header:
        return sorted(list(set(types)))

    
    # Check for number stuff
    if not is_id_header and should_check('numeric') and \
        do_check(is_a_number) and not is_boolean:
        
        types.append('numeric')
        
        # more number stuff
        if should_check('currency') and do_check(is_a_currency):
            types.append('currency')

        elif should_check('percent') and do_check(is_a_percent):
            types.append('percent')

        elif should_check('latitude') and do_check(is_a_latitude):
            types.append('latitude')

        elif should_check('longitude') and do_check(is_a_longitude):
            types.append('longitude')

    else:
        # string stuff
        if should_check('email') and do_check(is_a_email):
            types += ['email', 'label']

        elif should_check('url') and do_check(is_a_url):
            types += ['url']

        elif should_check('time') and do_check(is_a_time):
            types.append('time')

        elif should_check('coordinate') and do_check(is_a_coord_pair):
            types.append('coordinate')

        elif should_check('phone') and do_check(is_a_phone):
            types += ['phone']

        elif should_check('zip') and do_check(is_a_zip):
            types += ['zip']

        elif should_check('state') and do_check(is_a_state):
            types += ['state']

        elif should_check('country') and do_check(is_a_country):
            types += ['country']

        elif should_check('city') and do_check(is_a_city):
            types += ['city'] 

        elif should_check('address') and do_check(is_a_address):
            types.append('address')

        elif should_check('street') and do_check(is_a_street):
            types += ['street']

        elif should_check('label') and do_check(is_a_label):
            types += ['label']

    
    if len(types) == 0:
        types.append('unknown')

    return sorted(list(set(types)))



from faker import Factory
import numpy as np
import pandas as pd
import pprint
import json
from ColumnProcessor import ColType as ct
import ColumnProcessor as cp
import os.path
from random import randint
import random


class FakeDataDescriptor:
    def __init__(self):
        fake = Factory.create()
        fake.random.seed(1234)
        self.fake = fake
    methods = None
    with open(os.path.join(os.path.dirname(__file__),'types.json')) as data_file:
        methods = json.load(data_file)

    def show_example_data(self):
        for method in FakeDataDescriptor.methods:
            try:
                fake_func = getattr(self.fake, method)
                print("{0} - {1}.  Ex: {2}.... {3}".format(method,
                                                           FakeDataDescriptor.methods[method], str(fake_func()), str(fake_func())))
            except Exception as ex:
                print("{0} - Error: {1}".format(method, ex))


x = FakeDataDescriptor()
x.show_example_data()



# -*- coding: utf-8 -*-
"""
Created on Fri May 26 18:51:50 2017

@author: azunre
"""

import numpy as np

# standardize all column row number for fixed input-length CNN,
# used for standardization of encoded (3-dimensional) NN data
def DataLengthStandardizerEncoded(X, max_cells):
    nsamples, nx, ny = X.shape
    if nx >= max_cells:
        X = np.delete(X, np.arange(max_cells, nx), axis=1)  # truncate
    else:
        # replicate some randomly-selected cells
        X = np.concatenate(
            (X, X[:, np.random.choice(nx, max_cells - nx), :]), axis=1)
    return X
    
# standardize a single column row number (column as list input, array output, 
# used for column-wise data-length standardization of raw data)
def DataLengthColumnStandardizerRaw(column, max_cells):
    
    RAND = False # specify whether to pick a random subset of rows, or the first rows 
    
    if np.any(column[column.notnull()]):
        column = column[column.notnull()] # remove all nulls
    else:
        column = column.fillna('NaN')
    nrows = len(column)
    
    if nrows >= max_cells:
        column = np.asarray(column)
        if(RAND):
            column = column[np.random.choice(nrows,max_cells)] # random may not be preferred in this case
        else:
            column = np.delete(column,np.arange(max_cells,nrows))
    else:
        # replicate some randomly-selected cells
        column = np.asarray(column)
        column = np.concatenate([column,column[np.random.choice(nrows, max_cells - nrows)]])
        
    return column[np.newaxis].T

# standardize all column row number (used for whole raw dataset, input should
# a pandas dataframe, output also returns a flag that alerts system of inaccuracies
# for 'statistical' (ordinal, categorical, etc) variable prediction due to truncation)
def DataLengthStandardizerRaw(data, max_cells):
    nx, ny = data.shape
    out = DataLengthColumnStandardizerRaw(data.iloc[:,0],max_cells)
    for i in np.arange(1,ny):
        out = np.concatenate((out,DataLengthColumnStandardizerRaw(data.iloc[:,i],max_cells)),axis=1)
    return out



from distutils.core import setup

setup(name='d3m-simon',
    version='1.2.5',
    description='Character-level CNN+LSTM model for text classification',
    packages=['Simon', 'Simon.penny'],
    license='MIT',
    url='https://gitlab.com/datadrivendiscovery/contrib/simon',
    install_requires=['Faker >= 0.7.7',
        'scikit-learn==0.22.2.post1',
        'python-dateutil>=2.8.1',
        'pandas==1.1.3',
        'scipy==1.4.1',
        'tensorflow-gpu==2.2.0',
        'h5py >= 2.7.0'],
    include_package_data=True,
)




import numpy as np
from Simon.DataGenerator import *
# reduces down a a x b matrix to (a*b) x 2 matrix where the second column
# is the column names from the header


class DataFlattener:

    @staticmethod
    def to_flat(data, header):

        flat = np.empty((data.shape[0] * data.shape[1], 2))
        for row_num in range(0, data.shape[0]):
            for col_num in range(0,  data.shape[1]):
                flat[row_num * col_num + col_num, 0] = data[row_num, col_num]
                flat[row_num * col_num + col_num, 1] = header[col_num, 1]

        return np.matrix(flat)

    @staticmethod
    def flatten(data, header):
        rows = data.shape[0]
        cols = data.shape[1]
        return data.reshape(rows * cols).tolist(), np.tile(header, rows).tolist()
    #(X_train, y_train), (X_test, y_test)

    @staticmethod
    def get_flat_data(data, header):

        # data_matrix = pd.DataFrame(data, columns=header)
        # print data_matrix
        # data_matrix.to_csv('test_data_{}.csv'.format(data_matrix.shape[0]))
        x_data, y_data = DataGenerator.flatten(data, header)
        size = len(x_data)

        #np.save('test_data_{}.csv'.format(size), np.hstack((x_data,y_data)))

        train = int(np.ceil(size * 2 / 3))
        test = int(np.ceil(size / 3))
        return (x_data[:train], y_data[:train]), (x_data[train + 1:], y_data[train + 1:])
